(module app mzscheme
  (require "redex.scm"
           (lib "list.ss"))
    
  (define lang
    (language
     (p (store sf ...) e)
     (sf (x v))
     (e (e e ...) (set! x e) (begin e e ...) (if e e e) x v (e .. (mark e) e ...))
     (v (lambda (x ...) e) number #t #f - unspecified)
     (x (variable-except - lambda mark set! begin unspecified if))
     
     (P (store (sf ...) E))
     (E (e ... (mark E) e ...) (set! variable E) (begin E e e ...) (if E e e) hole)))
  
  (define v? (test-match lang v))
  
  (define reductions
    (reduction-relation
     lang
     ; application
     
     (--> (store ((x_s v_s) ...) (in-hole E_1 ((lambda (x_1 ...) e_body) v_1 ...)))
          (store ((x_s v_s) ... (x v_1) ...) (in-hole E_1 (app-subst ((x_1 ...) (x ...) e_body))))
          (side-condition
           (= (length (term (x_1 ...)))
              (length (term (v_1 ...)))))
          (fresh ((x ...) (x_1 ...))))
     
     ; deref
     (--> (store ((x_1 v_1) ... (x_i v_i) (x_2 v_2) ...) (in-hole E_1 (set! x_i v_ip)))
          (store ((x_1 v_1) ... (x_i v_ip) (x_2 v_2) ...) (in-hole E_1 unspecified)))
     
     ; set!
     (--> (store ((x_1 v_1) ... (x_i v_i) (x_2 v_2) ...) (in-hole E_1 x_i))
          (store ((x_1 v_1) ... (x_i v_i) (x_2 v_2) ...) (in-hole E_1 v_i)))
     
     ; if
     (--> (in-hole P_1 (if v_1 e_1 e_2)) 
          (in-hole P_1 e_1)
          (side-condition (not (equal? (term v_1) #f))))
     (--> (in-hole P_1 (if #f e_1 e_2)) 
          (in-hole P_1 e_2))
     
     ; begin
     (--> (in-hole P_1 (begin e_1))
          (in-hole P_1 e_1))
     (--> (in-hole P_1 (begin v_1 e_2 e_3 ...))
          (in-hole P_1 (begin e_2 e_3 ...)))
     
     ; negation
     (--> (in-hole P_1 (- number_1))
          (in-hole P_1 ,(- (term number_1))))
     
     ; mark introduction
     (--> (in-hole P_1 (e_1 ... e_i e_i+1 ...))
          (in-hole P_1 (e_1 ... (mark e_i) e_i+1 ...))
          (side-condition (not (v? (term e_i)))))
     
     ; mark removal
     (--> (in-hole P_1 (e_1 ... (mark v_i) e_i+1 ...))
          (in-hole P_1 (e_1 ... v_i e_i+1 ...)))))
  
  (define-metafunction app-subst
    lang
    [(any_1 any_2 any_3) ,(substitute-all (term any_1) (term any_2) (term any_3))])
  
  
  (define (substitute-all xs vs body)
    (foldl substitute body xs vs))
  
  (define substitute
    (plt-subst
     ['- (constant)]
     [(? symbol?) (variable)]
     [(? number?) (constant)]
     [`(lambda (,x ...) ,b)
      (all-vars x)
      (build (lambda (vars body) `(lambda ,vars ,body)))
      (subterm x b)]
     [`(set! ,x ,e)
      (all-vars '())
      (build (lambda (vars name body) `(set! ,name ,body)))
      (subterm '() x)
      (subterm '() e)]
     [`(if ,e1 ,e2 ,e3)
      (all-vars '())
      (build (lambda (vars e1 e2 e3) `(if ,e1 ,e2 ,e3)))
      (subterm '() e1)
      (subterm '() e2)
      (subterm '() e3)]
     [`(begin ,e ,e2 ...)
      (all-vars '())
      (build (lambda (vars e1 e2) `(begin ,e1 ,@e2)))
      (subterm '() e)
      (subterms '() e2)]
     [`(,f ,x ...)
      (all-vars '())
      (build (lambda (vars f . x) `(,f ,@x)))
      (subterm '() f)
      (subterms '() x)]))
  
  (provide reductions lang))
