#|

Note: the expression (lambda () (+)) is turned into just #%+ when used
in the r5 test suite, meaning that it is already a value and thus
doesn't contribute to the blowup that you get with marking non-values
in applications (say of dynamic wind). Accordingly, that expression
is used when there is a thunk in a dynamic wind that has no effects,
and whose result is ignored.

|#

(module dw-tests mzscheme
  (require "test.scm"
           (lib "match.ss"))
    
  (define (to-mz term)
    (match term
      [`(store ,binds (dw () ,e))
        `(let ,binds ,e)]))
  
  (define (conv-base base vec)
    (let loop ([i (vector-length vec)]
               [acc 0])
      (cond
        [(zero? i) acc]
        [else (loop (- i 1)
                    (+ acc (* (expt base (- i 1))
                              (vector-ref vec (- i 1)))))])))
  
  (define (deconv-base base number)
    (list->vector
     (let loop ([i number])
       (cond
         [(zero? i) '()]
         [else (cons (modulo i base)
                     (loop (quotient i base)))]))))
  
  (define dw-ts
    (test-suite
     "dw.scm"
     to-mz
     equal?
     
     (test "" '(store ([x 2]) (dw () x)) '(store ((x 2)) (dw () 2)))
     (test "" '(store ([x 2]) (dw () (begin (set! x (+ x 1)) x))) '(store ((x 3)) (dw () 3)))

     (test "" '(store () (dw () (begin ((lambda (x) (+ x x)) 1) 2))) '(store () (dw () 2)))
     
     (test ""
           '(store () (dw () (+ (call/cc (lambda (k) (+ (k 1) 1))) 1)))
           '(store () (dw () 2)))
     (test ""
           '(store () (dw () ((call/cc (lambda (x) x)) (lambda (y) 1))))
           '(store () (dw () 1)))
     
     (test ""
           '(store ((x 0))
                   (dw ()
                       (begin
                         (dynamic-wind (lambda () (set! x 1))
                                       (lambda () (set! x 2))
                                       (lambda () (set! x 3)))
                         x)))
           '(store ((x 3)) (dw () 3)))
     
     
     (test ""
           '(store ((x 0)) (dw () (dynamic-wind (lambda () (set! x (+ x 1)))
                                                (lambda () (begin (set! x (+ x 1)) x))
                                                (lambda () (set! x (+ x 1))))))
           '(store ((x 3)) (dw () 2)))
     
     (test "in thunk isn't really in"
           '(store ((n 0))
                   (dw ()
                       (begin
                         (call/cc
                          (lambda (k)
                            (dynamic-wind
                             (lambda () (begin
                                          (set! n (+ n 1))
                                          (k 11)))
                             (lambda () (+))
                             (lambda () (set! n (+ n 1))))))
                         n)))
           '(store ((n 1)) (dw () 1)))
     
     (test "out thunk is really out"
           '(store ((n 0)
                    (do-jump? #t)
                    (k-out #f))
                   (dw ()
                       (begin
                         (call/cc
                          (lambda (k)
                            (dynamic-wind
                             (lambda () (set! n (+ n 1)))
                             (lambda () (+))
                             (lambda ()
                               (begin
                                 (set! n (+ n 1))
                                 (call/cc (lambda (k) (set! k-out k))))))))
                         (if do-jump?
                             (begin
                               (set! do-jump? #f)
                               (k-out 0))
                             11)
                         (set! k-out #f)
                         n)))
           '(store ((n 2) (do-jump? #f) (k-out #f)) (dw () 2)))
     
     (test "out thunk is really out during trimming"
           '(store ((n 0)
                    (do-jump? #t)
                    (k-out #f))
                   (dw ()
                       (begin
                         (call/cc
                          (lambda (k)
                            (dynamic-wind
                             (lambda () (set! n (+ n 1)))
                             (lambda () (+))
                             (lambda ()
                               (begin
                                 (set! n (+ n 1))
                                 (call/cc (lambda (k) (set! k-out k))))))))
                         (if do-jump?
                             (begin
                               (set! do-jump? #f)
                               (k-out 0))
                             11)
                         (set! k-out #f)
                         n)))
           '(store ((n 2) (do-jump? #f) (k-out #f)) (dw () 2)))
     
     (test "jumping during the results of trimming, pre-thunk"
           '(store ((pre-count 0)
                    (pre-jump? #f)
                    (after-jump? #t)
                    (grab? #t)
                    (the-k #f))
                   (dw ()
                       (begin
                         (dynamic-wind
                          (lambda ()
                            (begin
                              (set! pre-count (+ pre-count 1))
                              (if pre-jump?
                                  (begin
                                    (set! pre-jump? #f)
                                    (set! after-jump? #f)
                                    (the-k 999))
                                  999)))
                          (lambda () 
                            (if grab?
                                (call/cc
                                 (lambda (k)
                                   (begin
                                     (set! grab? #f)
                                     (set! the-k k))))
                                999))
                          (lambda () (+)))
                         (if after-jump?
                             (begin
                               (set! pre-jump? #t)
                               (the-k 999))
                             999)
                         (set! the-k #f) ;; just to make testing simpler
                         pre-count)))
           '(store ((pre-count 3) (pre-jump? #f) (after-jump? #f) (grab? #f) (the-k #f)) (dw () 3)))
     
     (test "jumping during the results of trimming, post-thunk"
           '(store ((post-count 0)
                    (post-jump? #t)
                    (jump-main? #t)
                    (grab? #t)
                    (the-k #f))
                   (dw ()
                       (begin
                         (if grab?
                             (call/cc
                              (lambda (k)
                                (begin
                                  (set! grab? #f)
                                  (set! the-k k))))
                             999)
                         (dynamic-wind
                          (lambda () (+))
                          (lambda () 
                            (if jump-main?
                                (begin
                                  (set! jump-main? #f)
                                  (the-k 999))
                                999))
                          (lambda () 
                            (begin
                              (set! post-count (+ post-count 1))
                              (if post-jump?
                                  (begin
                                    (set! post-jump? #f)
                                    (the-k 999))
                                  999))))
                         (set! the-k #f) ;; just to make testing simpler
                         post-count)))
           '(store ((post-count 2) (post-jump? #f) (jump-main? #f) (grab? #f) (the-k #f)) (dw () 2)))

     
#|

to read the following tests, read the argument to conv-base from right to left
each corresponding set! should happen in that order.
in case of a test case failure, turn the number back into a sequence
of digits with deconv-base

these tests are written in a funny way to avoid a huge blowup 
(due to the order of evaluation issues) when running in the
r5 sematics

|#
     
     (test "hop out one level"
           '(store ((x 0)
                    (one 0)
                    (two 0)
                    (three 0)) 
                   (dw () 
                       (begin
                         (set! one (lambda () (set! x (+ (* x 2) 0))))
                         (set! two (lambda () (call/cc (lambda (k) k))))
                         (set! three (lambda () (set! x (+ (* x 2) 1))))
                         ((dynamic-wind one two three)
                          (lambda (y) x)))))
           (let ([final-x (conv-base 2 #(1 0 1 0))])
             `(store ((x ,final-x)
                      (one (lambda () (set! x (+ (* x 2) 0))))
                      (two (lambda () (call/cc (lambda (k) k))))
                      (three (lambda () (set! x (+ (* x 2) 1)))))
                     (dw () ,final-x))))
     
     (test "hop out two levels"
           '(store ((x 0)
                    (one 0)
                    (two 0)
                    (three 0)
                    (four 0))
                   (dw () 
                       (begin
                         (set! one   (lambda () (set! x (+ (* x 5) 1))))
                         (set! two   (lambda () (set! x (+ (* x 5) 2))))
                         (set! three (lambda () (set! x (+ (* x 5) 3))))
                         (set! four  (lambda () (set! x (+ (* x 5) 4))))
                         ((dynamic-wind 
                           one
                           (lambda () 
                             (dynamic-wind
                              two
                              (lambda () (call/cc (lambda (k) k)))
                              three))
                           four)
                          (lambda (y) x)))))
           (let ([final-x (conv-base 5 #(4 3 2 1 4 3 2 1))])
             `(store ((x ,final-x)
                      (one   (lambda () (set! x (+ (* x 5) 1))))
                      (two   (lambda () (set! x (+ (* x 5) 2))))
                      (three (lambda () (set! x (+ (* x 5) 3))))
                      (four  (lambda () (set! x (+ (* x 5) 4)))))
                     (dw () ,final-x))))

     (test "don't duplicate tail"
           '(store ((x 0)
                    (one 0)
                    (two 0)
                    (three 0)
                    (four 0))
                   (dw () 
                       (begin
                         (set! one (lambda () (set! x (+ (* x 5) 1))))
                         (set! two (lambda () (set! x (+ (* x 5) 2))))
                         (set! three (lambda () (set! x (+ (* x 5) 3))))
                         (set! four (lambda () (set! x (+ (* x 5) 4))))
                         (dynamic-wind 
                          one
                          (lambda () 
                            ((dynamic-wind two
                                           (lambda () (call/cc (lambda (k) k)))
                                           three)
                             (lambda (y) x)))
                          four))))
           `(store ((x ,(conv-base 5 #(4 3 2 3 2 1)))
                    (one (lambda () (set! x (+ (* x 5) 1))))
                    (two (lambda () (set! x (+ (* x 5) 2))))
                    (three (lambda () (set! x (+ (* x 5) 3))))
                    (four (lambda () (set! x (+ (* x 5) 4)))))
                   (dw () 
                       ,(conv-base 5 #(3 2 3 2 1)))))
           
     (test "dont' duplicate tail, 2 deep"
           '(store ((x 0)
                    (one 0)
                    (two 0)
                    (three 0)
                    (four 0)
                    (five 0)
                    (six 0))
                   (dw () 
                       (begin
                         (set! one (lambda () (set! x (+ (* x 7) 1))))
                         (set! two (lambda () (set! x (+ (* x 7) 2))))
                         (set! three (lambda () (set! x (+ (* x 7) 3))))
                         (set! four (lambda () (set! x (+ (* x 7) 4))))
                         (set! five (lambda () (set! x (+ (* x 7) 5))))
                         (set! six (lambda () (set! x (+ (* x 7) 6))))
                         (dynamic-wind 
                          one
                          (lambda () 
                            (dynamic-wind 
                             two
                             (lambda () 
                               ((dynamic-wind three
                                              (lambda () (call/cc (lambda (k) k)))
                                              four)
                                (lambda (y) x)))
                             five))
                          six))))
           
           `(store ((x ,(conv-base 7 #(6 5 4 3 4 3 2 1)))
                    (one (lambda () (set! x (+ (* x 7) 1))))
                    (two (lambda () (set! x (+ (* x 7) 2))))
                    (three (lambda () (set! x (+ (* x 7) 3))))
                    (four (lambda () (set! x (+ (* x 7) 4))))
                    (five (lambda () (set! x (+ (* x 7) 5))))
                    (six (lambda () (set! x (+ (* x 7) 6)))))
                   (dw () ,(conv-base 7 #(4 3 4 3 2 1)))))
     
     (test "hop out and back into another one"
           '(store ((x 0)
                    (one 0)
                    (two 0)
                    (three 0)
                    (four 0))
                   (dw () 
                       (begin
                         (set! one (lambda () (set! x (+ (* x 5) 1))))
                         (set! two (lambda () (set! x (+ (* x 5) 2))))
                         (set! three (lambda () (set! x (+ (* x 5) 3))))
                         (set! four (lambda () (set! x (+ (* x 5) 4))))
                         ((lambda (ok)
                            (dynamic-wind one
                                          (lambda () (ok (lambda (y) x)))
                                          two))
                          (dynamic-wind three
                                        (lambda () (call/cc (lambda (k) k)))
                                        four)))))
           `(store ((x ,(conv-base 5 #(2 1 4 3 2 1 4 3)))
                    (one (lambda () (set! x (+ (* x 5) 1))))
                    (two (lambda () (set! x (+ (* x 5) 2))))
                    (three (lambda () (set! x (+ (* x 5) 3))))
                    (four (lambda () (set! x (+ (* x 5) 4)))))
                   (dw () ,(conv-base 5 #(1 4 3 2 1 4 3)))))
     
     (test "hop out one level and back in two levels"
           '(store ((x 0)
                    (one 0)
                    (two 0)
                    (three 0)
                    (four 0))
                   (dw () 
                       (begin
                         (set! one (lambda () (set! x (+ (* x 5) 1))))
                         (set! two (lambda () (set! x (+ (* x 5) 2))))
                         (set! three (lambda () (set! x (+ (* x 5) 3))))
                         (set! four (lambda () (set! x (+ (* x 5) 4))))
                         ((lambda (ok)
                            (dynamic-wind
                             one
                             (lambda ()
                               (dynamic-wind
                                two
                                (lambda () (ok (lambda (y) x)))
                                three))
                             four))
                          (call/cc (lambda (k) k))))))
           `(store ((x ,(conv-base 5 #(4 3 2 1 4 3 2 1)))
                    (one (lambda () (set! x (+ (* x 5) 1))))
                    (two (lambda () (set! x (+ (* x 5) 2))))
                    (three (lambda () (set! x (+ (* x 5) 3))))
                    (four (lambda () (set! x (+ (* x 5) 4)))))
                   (dw () ,(conv-base 5 #(2 1 4 3 2 1)))))
     
     (test "hop out two levels and back in two levels"
           '(store ((x 0)
                    (one 0)
                    (two 0)
                    (three 0)
                    (four 0)
                    (five 0)
                    (six 0)
                    (seven 0)
                    (eight 0))
                   (dw () 
                       (begin
                         (set! one (lambda () (set! x (+ (* x 9) 1))))
                         (set! two (lambda () (set! x (+ (* x 9) 2))))
                         (set! three (lambda () (set! x (+ (* x 9) 3))))
                         (set! four (lambda () (set! x (+ (* x 9) 4))))
                         (set! five (lambda () (set! x (+ (* x 9) 5))))
                         (set! six (lambda () (set! x (+ (* x 9) 6))))
                         (set! seven (lambda () (set! x (+ (* x 9) 7))))
                         (set! eight (lambda () (set! x (+ (* x 9) 8))))
                         ((lambda (ok)
                            (dynamic-wind
                             one
                             (lambda () 
                               (dynamic-wind
                                two
                                (lambda () (ok (lambda (y) x)))
                                three))
                             four))
                          (dynamic-wind
                           five
                           (lambda () 
                             (dynamic-wind
                              six
                              (lambda () (call/cc (lambda (k) k)))
                              seven))
                           eight)))))
           `(store ((x ,(conv-base 9 #(4 3 2 1 8 7 6 5 4 3 2 1 8 7 6 5)))
                    (one (lambda () (set! x (+ (* x 9) 1))))
                    (two (lambda () (set! x (+ (* x 9) 2))))
                    (three (lambda () (set! x (+ (* x 9) 3))))
                    (four (lambda () (set! x (+ (* x 9) 4))))
                    (five (lambda () (set! x (+ (* x 9) 5))))
                    (six (lambda () (set! x (+ (* x 9) 6))))
                    (seven (lambda () (set! x (+ (* x 9) 7))))
                    (eight (lambda () (set! x (+ (* x 9) 8)))))
                   (dw () ,(conv-base 9 #(2 1 8 7 6 5 4 3 2 1 8 7 6 5)))))))
  
  (provide dw-ts))
