(module eval mzscheme

  (require "redex.scm"
           "reify.scm"
           (prefix srfi1: (lib "1.ss" "srfi"))
           (lib "match.ss"))
  
  (define lang
    (language (p (store (sf ...) e))
              (sf (x v)
                  (ptr (cons v v)))
              (e (e e ...)
                 (set! x e)
                 (begin e e ...)
                 (if e e e)
                 x
                 v)
              (v (lambda (x ...) e)
                 sqv
                 unspecified
                 -

                 'sy
                 ptr
                 null
                 prim)
              
              (P (store (sf ...) E))
              (E (v ... E e ...)
                 (set! x E)
                 (begin E e e ...)
                 (if E e e)
                 hole)
              
              (s (s ...)
                 sqv
                 sy
                 (s ... dot sqv)
                 (s ... dot sy))
              (sqv number
                   #t
                   #f)
              
              ;; compilation context
              (SP (store (sf ...) S))
              (S hole (e ... S s ...)
                 (lambda (x ...) S)
                 (if S s s)
                 (if e S s)
                 (if e e S)
                 (ccons v S)
                 (ccons S s))
              
              (prim eval cons car cdr eqv?)
              
              (sy (variable-except dot))
              
              (x (side-condition (name x (variable-except 
                                          lambda dot loc    ; core syntax names
                                          if begin
                                          null set!
                                          -
                                          cons car cdr  ; list functions
                                          quote
                                          eqv?
                                          eval ccons))
                                 (not (prefixed-by? (term x) 'ptr))))
              
              ; pair pointer
              (ptr (side-condition variable_p (prefixed-by? (term variable_p) 'ptr)))))
  
  (define reductions
    (reduction-relation 
     lang
     
     ;; cons
     (--> (store (sf_1 ...) (in-hole E_1 (cons v_car v_cdr)))
          (store (sf_1 ... (ptr (cons v_car v_cdr)))
            (in-hole E_1 ptr))
          (fresh ptr)
          "ECons")
     
     ;; car
     (--> (store (sf_1 ... (ptr_i (cons v_car v_cdr)) sf_2 ...)
            (in-hole E_1 (car ptr_i)))
          (store (sf_1 ... (ptr_i (cons v_car v_cdr)) sf_2 ...)
            (in-hole E_1 v_car))
          "ECar")
     
     ;; cdr
     (--> (store (sf_1 ... (ptr_i (cons v_car v_cdr)) sf_2 ...)
            (in-hole E_1 (cdr ptr_i)))
          (store (sf_1 ... (ptr_i (cons v_car v_cdr)) sf_2 ...) (in-hole E_1 v_cdr))
          "ECdr")
     
     ;; eqv? #t
     (--> (in-hole P_1 (eqv? ptr_1 ptr_1)) 
          (in-hole P_1 #t)
          "EEqv1")
     
     ;; eqv #f
     (--> (in-hole P_1 (eqv? ptr_1 ptr_2))
          (in-hole P_1 #f)
          "EEqv2"
          (side-condition (not (equal? (term ptr_1) (term ptr_2)))))
     
     ;; compile time quote removal
     (--> (in-hole SP_1 '(s_1 s_2 ...)) 
          (in-hole SP_1 (ccons 's_1 '(s_2 ...)))
          "EQSeq")
     (--> (in-hole SP_1 '())
          (in-hole SP_1 null)
          "EQNull")
     (--> (in-hole SP_1 '(s_1 s_2 s_3 ... dot s_4))
          (in-hole SP_1 (ccons 's_1 '(s_2 s_3 ... dot s_4)))
          "EQSDot")
     (--> (in-hole SP_1 '(s_1 dot s_2))
          (in-hole SP_1 (ccons 's_1 's_2))
          "EQDot")
     (--> (in-hole SP_1 'sqv_1)
          (in-hole SP_1 sqv_1)
          "EQSqv")

     ;; compile time pair allocation
     (--> (store (sf_1 ...) (in-hole S_1 (ccons v_1 v_2)))
          (store (sf_1 ... (ptr (cons v_1 v_2))) 
            (in-hole S_1 ptr))
          (fresh ptr)
          "EQPair")
     
     ;; eval
     (--> (store (sf_1 ...) (in-hole E_1 (eval v_1)))
          (store (sf_1 ...) (in-hole E_1 ,((current-reify) (term (sf_1 ...)) (term v_1))))
          "EEval")

     
     ;; application
     (--> (store (sf_1 ...)
            (in-hole E_1 ((lambda (x_arg1 ...) e_body) v_arg1 ...)))
          
          (store (sf_1 ... (xargs v_arg1) ...)
            (in-hole E_1
                     ,(subst-all 
                       (term (x_arg1 ...)) 
                       (term (xargs ...))
                       (term e_body))))
          
          (side-condition
           (= (length (term (x_arg1 ...))) (length (term (v_arg1 ...)))))
          (fresh ((xargs ...) (x_arg1 ...)))
          "EApp")
    
    ;; var lookup
    (--> (store (sf_before ... (x_1 v_1) sf_after ...)
                (in-hole E_1 x_1))
         (store (sf_before ... (x_1 v_1) sf_after ...)
                (in-hole E_1 v_1))
         "EVar")
    
    ;; set!
    (--> (store (sf_before ... (x_1 v_1) sf_after ...)
                (in-hole E_1 (set! x_1 v_2)))
         (store (sf_before ... (x_1 v_2) sf_after ...)
                (in-hole E_1 unspecified))
         "ESet!")
    
    ;; begin
    (--> (in-hole P_1 (begin e_1))
         (in-hole P_1 e_1)
         "EBegin1")
    (--> (in-hole P_1 (begin v_1 e_1 e_2 ...))
         (in-hole P_1 (begin e_1 e_2 ...))
         "EBegin2")
    
    ;; subtraction
    (--> (in-hole P_1 (- number_1))
         (in-hole P_1 ,(- (term number_1)))
         "E-")
    
    ;; if
    (--> (side-condition (in-hole P_1 (if v_1 e_1 e_2))
                         (not (eq? #f (term v_1))))
         (in-hole P_1 e_1)
         "EIf1")
    
    (--> (in-hole P_1 (if #f e_1 e_2))
         (in-hole P_1 e_2)
         "EIf2")))

  (define subst-one 
    (subst
     [`cons (constant)]
     [`null (constant)]
     [`call/cc (constant)]
     [`mark (constant)]
     [`cwv-mark (constant)]
     [(? symbol?) (variable)]
     [(? number?) (constant)]
     [`(lambda ,(xs ... 'dot last) ,b)
       (all-vars (cons last xs))
       (build (lambda (vars body) `(lambda ,(xs 'dot last) ,body)))
       (subterm xs b)]
     [`(lambda ,(xs ...) ,b)
       (all-vars xs)
       (build (lambda (vars body) `(lambda ,xs ,body)))
       (subterm xs b)]
     [`(lambda ,x ,b)
       (all-vars (list x))
       (build (lambda (vars body) `(lambda ,x ,body)))
       (subterm x b)]
     [`(set! ,x ,e)
       (all-vars '())
       (build (lambda (vars x exp) `(set! ,x ,exp)))
       (subterm '() x)
       (subterm '() e)]
     [(f args ...)
      (all-vars '())
      (build (lambda (vars f . args) `(,f ,@args)))
      (subterm '() f)
      (subterms '() args)]))
  
  (define (subst-all params args body) (srfi1:fold subst-one body params args))
  
  (define (prefixed-by? s prefix) (and (regexp-match (format "^~a" prefix) (symbol->string s)) #t))
  
  (provide reductions
           lang))
