(module r5rs mzscheme
  (require (planet "reduction-semantics.ss" ("robby" "redex.plt" 2 4))
           (planet "subst.ss" ("robby" "redex.plt" 2 4)))
  
  (require (prefix srfi1: (lib "1.ss" "srfi"))
           
           (lib "plt-match.ss")
           (lib "list.ss")
           "trim.scm")
  
  (provide lang 
           reductions
           proof-reductions
           non-proof-reductions
           r5rs-subst-all)
  
  (define lang
    (language (p (store (sf ...)
                   (dw (dws ...)
                       d
                       d ...)))
              
              (sf (x v)
                  (pp (#%cons v v))
                  (cp (lambda (x ...) e e ...))
                  (mp (lambda (x x ... dot x) e e ...))
                  (mp (lambda x e e ...)))
              (dws (x (v) (v)))
              
              (d (define x e)
                 (beginD d ...)
                 e)
              
              (e (e e ...)
                 (e ... (mark e) e ...)
                 (if e e e)
                 (if e e)
                 (set! x e) 
                 (begin e e ...)
                 x
                 v
                 (push dws)
                 (pop)
                 (throw x dws ... (in-hole D x))
                 (lambda (x ...) e e ...)
                 (lambda (x x ... dot x) e e ...)
                 (lambda x e e ...))
              
              (v nonfun fun)
              
              (nonfun pp
                      #%null
                      'sym
                      sqv
                      unspecified)
              
              (fun ufun
                   aop     ; primitive functions
                   fun1
                   fun2
                   #%list
                   #%dynamic-wind
                   #%apply
                   #%values)
              (fun1 #%null? #%pair? #%car #%cdr #%call/cc #%eval)
              (fun2 #%cons #%set-car! #%set-cdr! #%eqv? #%call-with-values)
              (ufun cp mp) ; user functions
              
              (aop #%+ #%- #%/ #%*)
              
              
              (P (store (sf ...) W))
              (W (dw (dws ...) D d ...))
              (D E*
                 (define x Eo))
              (E hole
                 (e ... (mark Eo) e ...)
                 (if Eo e e)
                 (if Eo e)
                 (set! x Eo)
                 (begin E* e e ...)
                 (#%call-with-values (lambda () E* e ...) v))
              (Eo (hole single) E)
              (E* (hole multi) E)
              
              (ds es
                  (define x es)
                  (beginD ds ...))
              (es 's 
                  (ccons es es)
                  (es es ...)
                  (es ... (mark es) es ...)
                  (if es es es)
                  (if es es)
                  (set! x es) 
                  (begin es es ...)
                  x
                  v
                  (push dws)
                  (pop)
                  (throw x dws ... (in-hole D x))
                  (lambda (x ...) es es ...)
                  (lambda (x x ... dot x) es es ...)
                  (lambda x es es ...))
              
              (s (s ...)
                 (s ... s dot sqv)
                 (s ... s dot sym)
                 sqv
                 sym)
              (sqv n #t #f)
              (SP (store (sf ...) (dw (dws ...) d ... SD s ...)))
              (SD S
                  (define x S)
                  (beginD d ... SD s ...))
              (S hole
                 (e ... S s ...)
                 (if e e S)
                 (if e S s)
                 (if S s s)
                 (if e S)
                 (if S s)
                 (set! x S)
                 (begin e e ... S s ...)
                 (begin S s ...)
                 (throw x dws ... S)
                 (push (x e S))
                 (push (x S s))
                 (lambda (x ...) S s ...)
                 (lambda (x ...) e e ... S s ...)
                 (lambda (x x ... dot x) S s ...)
                 (lambda (x x ... dot x) e e ... S s ...)
                 (lambda x S s ...)
                 (lambda x e e ... S s ...)
                 (ccons v S)
                 (ccons S s)
                 (mark S))
              
              (sym (variable-except dot))
              
              (x (side-condition 
                  (name var_none
                        (variable-except 
                         dot                         ; the . in dotted pairs
                         lambda if loc set!          ; core syntax names
                         quote begin
                         
                         ccons                       ; compile-time cons
                         mark                        ; marked expressions (superscript o)
                         
                         #%null                      ; non-function values
                         unspecified
                         pair closure
                         
                         error                       ; signal an error
                         
                         beginD define               ; top-level stuff
                         
                         #%cons #%pair? #%null? #%car #%cdr       ; list functions
                         #%set-car! #%set-cdr! #%list
                         #%+ #%- #%* #%/                          ; math functions
                         #%call/cc throw push pop #%dynamic-wind  ; call/cc functions
                         #%values #%call-with-values              ; values functions
                         #%apply #%eqv? #%eval))
                  (and (not (prefixed-by? (term var_none) 'c:))
                       (not (prefixed-by? (term var_none) 'p:))
                       (not (prefixed-by? (term var_none) 'm:)))))
              
              ; pair pointer
              (pp (side-condition variable_pp (prefixed-by? (term variable_pp) 'p:)))
              ; pointer to a lambda function
              (cp (side-condition variable_pp (prefixed-by? (term variable_pp) 'c:)))
              ; pointer to a mu-lambda function
              (mp (side-condition variable_mp (prefixed-by? (term variable_mp) 'm:)))
              
              (n number)
              
              (p* (store (sf ...) (dw (dws ...) ds ds ...)))
              (a* (store (sf ...) (dw (dws ...) (#%values v ...)))
                  (error string))))
  
  ;; --> : one-step reduction, full term to full term. 
  (define-syntax (--> stx)
    (syntax-case stx ()
      [(_ t result name) 
       #'(reduction/name name lang t (term result))]))
  
  ;; /-> : one-step reduction, P[term] to P[term]
  (define-syntax (/-> stx)
    (syntax-case stx ()
      [(_ t result name)
       #'(reduction/context/name name lang P t (term result))]))
  
  (define Basic--syntactic--forms
    (list
     
     ;; if
     (/-> (side-condition (if v_1 e_1 e_2) (not (eq? (term v_1) #f)))
          e_1
          "5if3t")
     (/-> (if #f e_1 e_2)
          e_2
          "5if3f")
     
     (/-> (side-condition (if v_1 e_1) (not (eq? (term v_1) #f)))
          e_1
          "5if2t")
     (/-> (if #f e_1)
          unspecified
          "5if2f")
     
     ;; begin
     (/-> (begin (#%values v ...) e_1 e_2 ...)
          (begin e_1 e_2 ...)
          "5beginc")
     (/-> (begin e_1)
          e_1
          "5beginl")))
  
  (define Arithmetic
    (list
     
     ;; primitives for numbers
     (/-> (#%+) 0 "5+0")
     (/-> (#%+ n_1 n_2 ...) ,(sum-of (term (n_1 n_2 ...))) "5+")
     
     (/-> (#%- n_1) ,(- (term n_1)) "5u-")
     (/-> (#%- n_1 n_2 n_3 ...) 
          ,(- (term n_1) (sum-of (term (n_2 n_3 ...))))
          "5-")
     
     (/-> (#%*) 1 "5*1")
     (/-> (#%* n_1 n_2 ...) ,(product-of (term (n_1 n_2 ...))) "5*")
     
     (/-> (#%/ n_1) (#%/ 1 n_1)
          "5u/")
     (/-> (side-condition (#%/ n_1 n_2 n_3 ...) 
                          (not (member 0 (term (n_2 n_3 ...)))))
          ,(/ (term n_1) (product-of (term (n_2 n_3 ...))))
          "5/")))
  
  (define (sum-of x) (apply + x))
  (define (product-of x) (apply * x))
  
  (define Cons--and--cons-cell--mutation
    (list
     ;; #%cons
     (--> (name prog
                (store (sf_1 ...)
                  (in-hole W_1 (#%cons v_1 v_2))))
          ,(term-let ((pp_i (variable-not-in (term prog) 'p:)))
             (term
              (store (sf_1 ... (pp_i (#%cons v_1 v_2)))
                (in-hole W_1 pp_i))))
          "5cons")
     
     ;; #%list
     (/-> (#%list v_1 v_2 ...) (#%cons v_1 (#%list v_2 ...)) "5listc")
     (/-> (#%list) #%null "5listn")
     
     ;; #%car
     ((store (sf_1
              ...
              (pp_i (#%cons v_1 v_2))
              sf_2 ...)
        (in-hole W_1 (#%car pp_i)))
      . --> .
      (store (sf_1 
              ... 
              (pp_i (#%cons v_1 v_2)) 
              sf_2 ...)
        (in-hole W_1 v_1))
      "5car")
     
     ;; #%cdr
     ((store (sf_1
              ...
              (pp_i (#%cons v_1 v_2))
              sf_2 ...)
        (in-hole W_1 (#%cdr pp_i)))
      . --> .
      (store (sf_1 
              ... 
              (pp_i (#%cons v_1 v_2))
              sf_2 ...)
        (in-hole W_1 v_2))
      "5cdr")
     
     ;; #%null?
     ((#%null? #%null) . /-> . #t "5null?t")
     (/-> (side-condition (#%null? v_1) (not (null-v? (term v_1))))
          #f
          "5null?f")
     
     ;; #%pair? 
     ((#%pair? pp)  . /-> . #t "5pair?t")
     (/-> (side-condition (#%pair? v_1) (not (prefixed-by? (term v_1) 'p:)))
          #f
          "5pair?f")
     
     ;; #%set-car!
     ((store (sf_1
              ...
              (pp_1 (#%cons v_1 v_2))
              sf_2 ...)
        (in-hole W_1 (#%set-car! pp_1 v_3)))
      . --> .
      (store (sf_1 
              ... 
              (pp_1 (#%cons v_3 v_2))
              sf_2 ...)
        (in-hole W_1 unspecified))
      "5setcar")
     
     ;; #%set-cdr!
     ((store (sf_1
              ...
              (pp_1 (#%cons v_1 v_2))
              sf_2 ...)
        (in-hole W_1 (#%set-cdr! pp_1 v_3)))
      . --> .
      (store (sf_1 
              ... 
              (pp_1 (#%cons v_1 v_3))
              sf_2 ...)
        (in-hole W_1 unspecified))
      "5setcdr")))
  
  (define Procedure--application
    (list
     ;; closure introduction
     (--> (name prog (store (sf_1 ...) (in-hole W_1 (lambda (x_1 ...) e_1 e_2 ...))))
          ,(term-let ((cp_none (variable-not-in (term prog) 'c:)))
             (term (store (sf_1 ... (cp_none (lambda (x_1 ...) e_1 e_2 ...)))
                     (in-hole W_1 cp_none))))
          "5calloc")
     
     ;; mu-closure introduction
     ((name prog 
            (store (sf_1 ...)
              (in-hole W_1 (lambda (x_1 x_2 ... dot x_r) e_1 e_2 ...))))
      . --> .
      ,(term-let ((mp_none (variable-not-in (term prog) 'm:))
                  (cp_none (variable-not-in (term prog) 'c:)))
         (term 
          (store (sf_1 
                  ... 
                  (mp_none (lambda (x_1 x_2 ... dot x_r) (cp_none x_1 x_2 ... x_r)))
                  (cp_none (lambda (x_1 x_2 ... x_r) e_1 e_2 ...)))
            (in-hole W_1 mp_none))))
      "5μcalloc")
     
     ((name prog 
            (store (sf_1 ...)
              (in-hole W_1 (lambda x_1 e_1 e_2 ...))))
      . --> .
      ,(term-let ((mp_none (variable-not-in (term prog) 'm:))
                  (cp_none (variable-not-in (term prog) 'c:)))
         (term 
          (store (sf_1 
                  ... 
                  (mp_none (lambda x_1 (cp_none x_1)))
                  (cp_none (lambda (x_1) e_1 e_2 ...)))
            (in-hole W_1 mp_none))))
      "5μcalloc1")
     
     (--> (side-condition 
           (in-hole P_1 (e_1 ... e_i e_i+1 ...))
           (and (not (v? (term e_i)))
                (not (cwv-app? (term (e_1 ...))
                               (term e_i)
                               (term (e_i+1 ...))))))
          (in-hole P_1 (e_1 ... (mark e_i) e_i+1 ...))
          "5mark")
     
     ;; drop marks around values
     ((e_1 ... (mark v_i) e_i+1 ...)
      . /-> . 
      (e_1 ... v_i e_i+1 ...)
      "5unmark")
     
     ;; lambda application
     ((side-condition
       (name prog
             (store (sf_1 
                     ... 
                     (cp_i (lambda (x_1 ...) e_body1 e_body2 ...))
                     sf_2  ...)
               (in-hole W_1 (cp_i v_1 ...))))
       (= (length (term (x_1 ...))) (length (term (v_1 ...)))))
      . --> .
      ,(term-let ([(x_2 ...) 
                   (variables-not-in 
                    (term (v_1 ...))
                    (term prog)
                    'b:)])
         (term 
          (store (sf_1 
                  ...
                  (cp_i (lambda (x_1 ...) e_body1 e_body2 ...))
                  sf_2 ...
                  (x_2 v_1) ...)
            (in-hole W_1
                     ,(r5rs-subst-all 
                       (term (x_1 ...)) 
                       (term (x_2 ...))
                       (term (begin e_body1 e_body2 ...)))))))
      "5app")
     
     ;; mu-lambda application
     
     ((side-condition
       (store (sf_1
               ... 
               (mp_i (lambda (x_1 x_2 ... dot x_r) (cp_1 x_1 x_2 ... x_r)))
               sf_2 ...)
         (in-hole W_1 (mp_i v_1 ... v_2 ...)))
       (= (length (term (v_1 ...))) (length (term (x_1 x_2 ...)))))
      . --> .
      (store (sf_1
              ...
              (mp_i (lambda (x_1 x_2 ... dot x_r) (cp_1 x_1 x_2 ... x_r)))
              sf_2 ...)
        (in-hole W_1 (cp_1 v_1 ... (#%list v_2 ...))))
      "5μapp")
     
     ((store (sf_1
              ... 
              (mp_i (lambda x_1 (cp_1 x_1)))
              sf_2 ...)
        (in-hole W_1 (mp_i v_1 ...)))
      . --> .
      (store (sf_1
              ...
              (mp_i (lambda x_1 (cp_1 x_1)))
              sf_2 ...)
        (in-hole W_1 (cp_1 (#%list v_1 ...))))
      "5μapp1")
     
     ;; #%apply
     ((store (sf_1
              ...
              (pp_i (#%cons v_2 v_3))
              sf_2 ...)
        (in-hole W_1 (#%apply fun_1 v_1 ... pp_i)))
      . --> .
      (store (sf_1
              ...
              (pp_i (#%cons v_2 v_3))
              sf_2 ...)
        (in-hole  W_1
                  (#%apply fun_1 v_1 ... v_2 v_3)))
      "5applyc")
     
     ((#%apply fun_1 v_1 ... #%null)
      . /-> .
      (fun_1 v_1 ...)
      "5applyn")))
  
  (define Call-cc--and--dynamic-wind
    (list
     
     (--> (side-condition
           (name prog 
                 (in-hole P_1
                          (#%dynamic-wind v_1 v_2 v_3)))
           (and (arity-zero? (term prog) (term v_1))
                (arity-zero? (term prog) (term v_2))
                (arity-zero? (term prog) (term v_3))))
          ,(term-let ([d (variable-not-in (term prog) 'd)]
                      [x (variable-not-in (term prog) 'x)]
                      [cp (variable-not-in (term prog) 'c:)]
                      [mp (variable-not-in (term prog) 'm:)])
             (term 
              (in-hole P_1
                       (begin (v_1)
                              (push (d (v_1) (v_3)))
                              (#%call-with-values v_2
                                                  (lambda x (pop) (v_3) (#%apply #%values x)))))))
          "5dw")
     
     ;; push dynamic context
     ((store (sf_1 ...) (dw (dws_1 ...) (in-hole D_1 (push dws_2)) d_1 ...))
      . --> .
      (store (sf_1 ...) (dw (dws_1 ... dws_2) (in-hole D_1 unspecified) d_1 ...))
      "5push")
     
     ;; pop dynamic context
     ((store (sf_1 ...) (dw (dws_1 ... dws_2) (in-hole D_1 (pop)) d_1 ...))
      . --> .
      (store (sf_1 ...) (dw (dws_1 ...) (in-hole D_1 unspecified) d_1 ...))
      "5pop")
     
     ;; #%call/cc
     ((name prog (store (sf_1 ...) 
                   (dw (dws_1 ...) 
                       (in-hole D_1 (#%call/cc v_1))
                       d_1 ...)))
      . --> .
      ,(term-let ([x1 (variable-not-in (term prog) 'x)]
                  [x2 (variable-not-in (term prog) 'k)])
         (term (store (sf_1 ...) 
                 (dw (dws_1 ...) 
                     (in-hole D_1
                              (v_1
                               (lambda x1
                                 (throw
                                  x2 
                                  dws_1 ...
                                  (in-hole D_1
                                           (begin x2 (#%apply #%values x1)))))))
                     d_1 ...))))
      "5callcc")
     
     ;; throw to a continuation
     ((store (sf_1 ...)
        (dw (dws_1 ...)
            (in-hole D_1 (throw x_2 dws_2 ... (in-hole D_2 (begin x_2 e_1))))
            d_1 ...))
      . --> .
      (store (sf_1 ...) 
        (dw (dws_1 ...) 
            (in-hole D_2 (begin ,@(trim (term (dws_1 ...))
                                        (term (dws_2 ...)))
                                e_1))
            d_1 ...))
      "5throw")))
  
  (define (arity-zero? prog v)
    (cond
      [(memq v '(#%+ #%* #%list #%values)) #t]
      [(or (mp? v) (cp? v))
       (match prog
         [`(store ,s ,whatever ...)
          (let ([pr (assoc v s)])
            (and pr
                 (let ([sv (cadr pr)])
                   (match sv
                     [`(lambda () ,e ...) #t]
                     [`(lambda ,(? symbol?) , e ...) #t]
                     [else #f]))))])]
      [else #f]))
  
  (define Multiple--values--and--call-with-values
    (list
     ;; values promotion
     ((in-named-hole multi P_1 v_1)
      . --> .
      (in-hole P_1 (#%values v_1))
      "5promote")
     
     ;; values demotion
     ((in-named-hole single P_1 (#%values v_1))
      . --> .
      (in-hole P_1 v_1)
      "5demote")
     
     ; resolving call-with-values statements
     ((in-hole P_1 (#%call-with-values (lambda () (#%values v_2 ...)) v_1))
      . --> .
      (in-hole P_1 (v_1 v_2 ...))
      "5cwvd")
     
     ((in-hole P_1 (#%call-with-values (lambda () (#%values v_1 ...) e_1 e_2 ...) v_2))
      . --> .
      (in-hole P_1 (#%call-with-values (lambda () e_1 e_2 ...) v_2))
      "5cwvc")
     
     ((in-hole P_1 (#%call-with-values v_1 v_2))
      . --> .
      (in-hole P_1 (#%call-with-values (lambda () (v_1)) v_2))
      "5cwvw")))
  
  (define Eqv--and--equivalence
    (list
     ((#%eqv? v_1 v_1) . /-> . #t "5eqt")
     
     (/-> (side-condition (#%eqv? v_1 v_2) (not (equal? (term v_1) (term v_2))))
          #f
          "5eqf")
     
     (/-> (side-condition (#%eqv? ufun_1 ufun_2) (not (equal? (term ufun_1) (term ufun_2))))
          #t
          "5eqproof")))
  
  (define Errors
    (list
     (--> (side-condition (in-hole P (#%/ n_1 n_2 n_3 ...))
                          (member 0 (term (n_2 n_3 ...))))
          (error "divison by zero")
          "5/0")
     
     (--> (side-condition (in-hole P (aop v_1 ... v_i v_i+1 ...))
                          (not (number? (term v_i))))
          (error ,(format "arith-op applied to non-number, arg ~a"
                          (+ (length (term (v_1 ...))) 1)))
          "5ae")
     
     (--> (side-condition 
           (in-hole P_1 (#%car v_i))
           (not (cons-v? (term v_i))))
          (error "can't take car of non-pair")
          "5care")
     
     (--> (side-condition
           (in-hole P_1 (#%cdr v_i))
           (not (cons-v? (term v_i))))
          (error "can't take cdr of non-pair")
          "5cdre")
     
     (--> (side-condition (in-hole P_1 (#%set-car! v_1 v_2)) (not (cons-v? (term v_1))))
          (error "can't set-car! on a non-pair")
          "5scare")
     
     ;; #%set-cdr! error
     (--> (side-condition (in-hole P_1 (#%set-cdr! v_1 v_2)) (not (cons-v? (term v_1))))
          (error "can't set-cdr! on a non-pair")
          "5scdre")
     
     ;; attempt to refer to a free variable (as yet undefined variable)
     (--> (side-condition (store (sf_1 ...)
                            (dw (dws ...)
                                (in-hole D x_1)
                                d ...))
                          (not (memq (term x_1) (map car (term (sf_1 ...))))))
          (error ,(format "reference to free identifier: ~a" (term x_1)))
          "5errvar")
     
     ;; attempt to set! an as-yet undefined variable
     (--> (side-condition (store (sf_1 ...)
                            (dw (dws ...)
                                (in-hole D (set! x_1 v))
                                d ...))
                          (not (memq (term x_1) (map car (term (sf_1 ...))))))
          (error ,(format "attempt to set! free identifier: ~a" (term x_1)))
          "5errset")
     
     ;; application of non-function
     (--> (in-hole P (nonfun v ...))
          (error "can't apply non-function")
          "5appe")
     
     (--> (side-condition
           (in-hole P (#%apply fun v_1 ... v_2))
           (not (list-v? (term v_2))))
          (error "apply's last argument non-list")
          "5applye")
     
     (--> (in-hole P (#%apply nonfun v ...))
          (error "can't apply non-function")
          "5applynf")
     
     (--> (side-condition (name prog (in-hole P (#%dynamic-wind v_1 v_2 v_3)))
                          (or (not (arity-zero? (term prog) (term v_1)))
                              (not (arity-zero? (term prog) (term v_2)))
                              (not (arity-zero? (term prog) (term v_3)))))
          (error "dynamic-wind expects arity 0 procs")
          "5dwerr")
     
     
     
     (--> (side-condition (in-named-hole single P (#%values v_1 ...))
                          (not (= (length (term (v_1 ...))) 1)))
          (error "context received wrong # of values")
          "5valerr")
     
     (--> (side-condition (store (sf_1 ...)
                            (in-hole W (#%eval v_1)))
                          (not (ds? (reify (term (sf_1 ...)) (term v_1)))))
          (error ,(format "malformed expression: ~s" (reify (term (sf_1 ...)) (term v_1))))
          "5evale")
     
     (--> (side-condition (store (sf_1 ...)
                            (in-hole W (#%eval v_1)))
                          (and (ds? (reify (term (sf_1 ...)) (term v_1))) 
                               (not (es? (reify (term (sf_1 ...)) (term v_1))))))
          (error "eval only takes expressions")
          "5evald")))
  
  (define Arity--checking
    (list
     ((side-condition
       (store (sf 
               ... 
               (cp_i (lambda (x_1 ...) e e ...))
               sf ...)
         (in-hole W (cp_i v_1 ...)))
       (not (= (length (term (x_1 ...)))
               (length (term (v_1 ...))))))
      . --> .
      (error "arity mismatch")
      "5arity")
     
     ;; mu-lambda too few arguments case
     ((side-condition
       (store (sf 
               ...
               (mp_i (lambda (x_1 x_2 ... dot x) (cp x ...)))
               sf ...)
         (in-hole W (mp_i v_1 ...)))
       (< (length (term (v_1 ...)))
          (length (term (x_1 x_2 ...)))))
      . --> .
      (error "arity mismatch")
      "5μarity")
     
     (--> (side-condition (in-hole P (fun1 v_1 ...)) (not (= (length (term (v_1 ...))) 1)))
          (error "arity mismatch")
          "51arity")
     (--> (side-condition (in-hole P (fun2 v_1 ...)) (not (= (length (term (v_1 ...))) 2)))
          (error "arity mismatch")
          "52arity")     
     (--> (side-condition (in-hole P (#%dynamic-wind v_1 ...))
                          (not (= (length (term (v_1 ...))) 3)))
          (error "arity mismatch")
          "5dwarity")
     
     (--> (in-hole P (#%-)) (error "arity mismatch") "5-arity")
     (--> (in-hole P (#%/)) (error "arity mismatch") "5/arity")
     (--> (in-hole P (#%apply)) (error "arity mismatch") "5apparity0")
     (--> (in-hole P (#%apply v)) (error "arity mismatch") "5apparity1")))
  
  (define (kons kar kdr)
    (cond
      [(or (null? kdr) (pair? kdr)) `(,kar ,@kdr)]
      [else `(,kar dot ,kdr)]))
  
  (define Quote--and--eval
    (list
     
     ;; compile time quote removal
     (--> (in-hole SP_1 '(s_1 s_2 ...))
          (in-hole SP_1 (ccons 's_1 '(s_2 ...)))
          "5qcons")
     
     ((in-hole SP_1 '())
      . --> . 
      (in-hole SP_1 #%null)
      "5qnull")
     
     ((in-hole SP_1 '(s_1 s_2 s_3 ... dot s_4))
      . --> . 
      (in-hole SP_1 (ccons 's_1 '(s_2 s_3 ... dot s_4)))
      "5qconsd")
     
     ((in-hole SP_1 '(s_1 dot s_2))
      . --> . 
      (in-hole SP_1 (ccons 's_1 's_2))
      "5qdot")
     
     ((in-hole SP_1 'sqv_1)
      . --> .
      (in-hole SP_1 sqv_1)
      "5qsqv")
     
     ;; compile time pair allocation
     ((name exp
            (store (sf_1 ...)
              (dw (dws_1 ...) 
                  d_1 ...
                  (in-hole SD_1 (ccons v_1 v_2))
                  s_1 ...)))
      . --> . 
      ,(term-let ([pp1 (variable-not-in (term exp) 'p:)])
         (term (store (sf_1 ... (pp1 (#%cons v_1 v_2)))
                 (dw (dws_1 ...)
                     d_1 ...
                     (in-hole SD_1 pp1)
                     s_1 ...))))
      "5ccons")
     
     
     ((side-condition (store (sf_1 ...)
                        (in-hole W_1 (#%eval v_1)))
                      (es? (reify (term (sf_1 ...)) (term v_1))))
      . --> . 
      (store (sf_1 ...)
        (in-hole W_1 ,(reify (term (sf_1 ...)) (term v_1))))
      "5eval")))
  
  (define (reify store v)
    (match v
      [`#%null '()]
      [`(quote ,(? symbol? s)) s]
      [(? pp?)
       (let ([ent (assq v store)])
         (unless ent (error 'reify "found dangling pointer ~s" v))
         (match ent 
           [`(,var (#%cons ,va ,vd))
            (kons (reify store va) (reify store vd))]
           [else (error 'reify "variable not bound to pair ~s, ~s" v ent)]))]
      [else v]))
  
  (define (well-formed? a b) (es? (reify a b)))
  
  
  (define Top--level--and--Variables
    (list
     
     ;; define a variable for the first time
     (--> (side-condition (store (sf_1 ...)
                            (dw (dws_1 ...)
                                (define x_1 v_1)
                                d_1 ...))
                          (not (memq (term x_1) (map car (term (sf_1 ...))))))
          (store (sf_1 ... (x_1 v_1))
            (dw (dws_1 ...) 
                unspecified
                d_1 ...))
          "5def")
     
     ;; re-define a variable
     (--> (store (sf_1 ... (x_1 v_2) sf_2 ...)
            (dw (dws_1 ...) 
                (define x_1 v_1) 
                d_1 ...))
          (store (sf_1 ... (x_1 v_1) sf_2 ...)
            (dw (dws_1 ...) 
                unspecified
                d_1 ...))
          "5redef")
     
     ;; splice out begin
     (--> (store (sf_1 ...) (dw (dws_1 ...) (beginD d_1 d_2 ...) d_3 ...))
          (store (sf_1 ...) (dw (dws_1 ...) d_1 d_2 ... d_3 ...))
          "5tbegin")
     
     ;; drop empty begin
     (--> (store (sf_1 ...) (dw (dws_1 ...) (beginD) d_1 ...))
          (store (sf_1 ...) (dw (dws_1 ...) unspecified d_1 ...))
          "5tbegine")
     
     ;; drop some values
     (--> (store (sf_1 ...) (dw (dws_1 ...) (#%values v ...) d_1 d_2 ...))
          (store (sf_1 ...) (dw (dws_1 ...) d_1 d_2 ...))
          "5tdrop")
     
     ;; variable lookup
     ((store (sf_1
              ... 
              (x_1 v_1)
              sf_2 ...)
        (in-hole W_1 x_1))
      . --> . 
      (store (sf_1 
              ...
              (x_1 v_1)
              sf_2 ...)
        (in-hole W_1 v_1))
      "5var")
     
     ;; set!
     ((store (sf_1 ... (x_1 v_1) sf_2 ...)
        (in-hole W_1 (set! x_1 v_2)))
      . --> .
      (store (sf_1 ... (x_1 v_2) sf_2 ...)
        (in-hole W_1 unspecified))
      "5set")))
  
  (define (in-store? store . names)
    (andmap (λ (x) (assq x store)) names))
  
  (define reductions (append Arithmetic
                             Basic--syntactic--forms
                             Cons--and--cons-cell--mutation
                             Procedure--application
                             Call-cc--and--dynamic-wind
                             Multiple--values--and--call-with-values
                             Eqv--and--equivalence
                             Quote--and--eval
                             Top--level--and--Variables
                             Errors
                             Arity--checking))
  
  (define proof-reductions (filter (λ (x) (regexp-match #rx"proof" (red-name x))) reductions))
  (define non-proof-reductions (filter (λ (x) (not (memq x proof-reductions))) reductions))
  
  ;; BROKEN!
  (define r5rs-subst-one 
    (subst
     [`#%cons (constant)]
     [`#%null (constant)]
     [`throw (constant)]
     [`#%call/cc (constant)]
     [`mark (constant)]
     [`(quote ,x) (constant)]
     [(? symbol?) (variable)]
     [(? number?) (constant)]
     [(? boolean?) (constant)]
     [`(lambda ,(xs ... 'dot last) ,@(bs ...))
      (all-vars (cons last xs))
      (build (lambda (vars body) `(lambda ,(xs 'dot last) ,body)))
      (subterms xs bs)]
     [`(lambda ,(xs ...) ,@(bs ...))
      (all-vars xs)
      (build (lambda (vars . bodies) `(lambda ,xs ,@bodies)))
      (subterms xs bs)]
     [`(set! ,x ,e)
      (all-vars '())
      (build (lambda (vars x exp) `(set! ,x ,exp)))
      (subterm '() x)
      (subterm '() e)]
     [(f args ...)
      (all-vars '())
      (build (lambda (vars f . args) `(,f ,@args)))
      (subterm '() f)
      (subterms '() args)]))
  
  (define (r5rs-subst-all params args body) (srfi1:fold r5rs-subst-one body params args))
  
  (define (variables-not-in items exp prefix)
    (cond
      [(null? items) null]
      [else 
       (let ((this (variable-not-in exp prefix)))
         (cons 
          this 
          (variables-not-in 
           (cdr items)
           (cons this exp)
           prefix)))]))
  
  (define v? (language->predicate lang 'v))
  (define cons-v? (language->predicate lang 'pp))
  (define (null-v? v) (eq? v '#%null))
  (define (list-v? v) (or (cons-v? v) (null-v? v)))
  (define mp? (language->predicate lang 'mp))
  (define pp? (language->predicate lang 'pp))
  (define cp? (language->predicate lang 'cp))
  (define es? (language->predicate lang 'es))
  (define ds? (language->predicate lang 'ds))
  
  (define hole-gensym (gensym 'hole))
  (define cwv-app?
    (let ([pat1 (compile-pattern lang `(#%call-with-values))]
          [pat2 (compile-pattern lang `(lambda () e))]
          [pat3 (compile-pattern lang '(e))])
      (λ (x1 x2 x3)
        (and (match-pattern pat1 x1)
             (match-pattern pat2 x2)
             (match-pattern pat3 x3)))))
  
  (define (prefixed-by? s prefix)
    (let* ([sym (symbol->string s)]
           [pre (symbol->string prefix)]
           [plen (string-length pre)]
           [slen (string-length sym)])
      (and (slen . >= . plen)
           (string=? (substring sym 0 plen) pre)))))
