(module show-examples mzscheme
  
  (require (planet "gui.ss" ("robby" "redex.plt" 3))
           (prefix app: "app.scm")
           (prefix app-bad: "app-bad.scm")
           (prefix eval: "eval.scm")
           (prefix mv: "mv.scm")
           (prefix dw: "dw.scm")
           (prefix tl: "tl.scm"))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;;  functions that show terms from each of the systems
  ;;
  
  (define (app-show x) (stepper app:lang app:reductions x))
  (define (app-bad-show x) (stepper app-bad:lang app-bad:reductions x))
  (define (eval-show x) (stepper eval:lang eval:reductions x))
  (define (mv-show x) (stepper mv:lang mv:reductions x))
  (define (dw-show x) (stepper dw:lang dw:reductions x))
  (define (tl-show x) (stepper tl:lang tl:reductions x))

  (provide app-show app-bad-show eval-show mv-show dw-show tl-show)
  
  (reduction-steps-cutoff 1000)

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;;  predicate that returns #t if a term
  ;;  imposes a proof obligation that two
  ;;  procedures are the same
  ;;
  
  (define (no-proof-pred sexp term-node)
    (let ([parents (term-node-parents term-node)])
      (when (null? parents)
        (definitely-no-proof term-node))
      (when (ormap (λ (parent) (hash-table-get definitely-no-proof-ht parent (λ () #f)))
                   (term-node-parents term-node))
        (unless (andmap (λ (parent-label) (regexp-match #rx"proof" parent-label))
                        (term-node-labels term-node))
          (definitely-no-proof term-node)))
      (unless (hash-table-get definitely-no-proof-ht term-node (λ () #f))
        (hash-table-put! currently-requires-proof term-node #t))
      (hash-table-get definitely-no-proof-ht term-node (λ () #f))))
  
  (define (definitely-no-proof term-node)
    (hash-table-put! definitely-no-proof-ht term-node #t)
    (when (hash-table-get currently-requires-proof term-node (λ () #f))
      (hash-table-remove! currently-requires-proof term-node)
      (for-each
       (λ (term-node)
         (let ([now-needs-proof? (needs-proof? term-node)])
           (unless now-needs-proof?
             (term-node-set-red! term-node #f)
             (hash-table-remove! currently-requires-proof term-node))))
       (hash-table-map currently-requires-proof (λ (term-node _) term-node)))))
  
  (define (needs-proof? term-node)
    (let ([parents (term-node-parents term-node)])
      (if (null? parents)
          #f
          (let ([parents-need-proof? (map needs-proof? parents)])
            (andmap (λ (parent-needs-proof? parent-label)
                      (or parent-needs-proof?
                          (and (not parent-needs-proof?)
                               (regexp-match #rx"proof" parent-label))))
                    parents-need-proof?
                    (term-node-labels term-node))))))
  
  (define definitely-no-proof-ht (make-hash-table))
  (define currently-requires-proof (make-hash-table))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;
  ;; example uses of the above functions
  ;; #; comments out an entire sexpression.
  ;;

  #;
  (app-show '(store ((x 1)) ((set! x (- x)) (set! x (- x)))))
  
  #;
  (app-bad-show '(store ((x 1)) ((set! x (- x)) (set! x (- x)))))
  
  #;
  (eval-show '(store () ((lambda (f) (eqv? (f) (f)))
                         (lambda () '(x)))))
  
  #;
  (mv-show '(store ()
                   ((lambda (x) (x x)) 
                    ((lambda (y) y) 
                     (apply-values (lambda (x) (values x))
                                   (lambda (x) (x x)))))))
  
  #;
  (dw-show '(store ((x 0))
                   (dw ()
                       ((dynamic-wind (lambda () (set! x (+ (* x 2) 0)))
                                      (lambda () (call/cc (lambda (k) k)))
                                      (lambda () (set! x (+ (* x 2) 1))))
                        (lambda (y) x)))))
  
  #;
  (tl-show '(store ()
                   (define x 1) 
                   (beginD 
                     (define k (call/cc (lambda (x) x))) 
                     (set! x (+ x 1)))
                   (k 1)))
  
  (eval-show '(store () (eval ''(a dot b))))
  )

