
(module tl mzscheme
  (require "redex.scm"
           (lib "match.ss"))
  
  (define lang
    (language
     (p (store ((x v) ...) d ...))
     (d e
        (define x e)
        (beginD d ...))
     (e (set! x e)
        x
        (e e ...)
        (begin e e ...)
        (if e e e)
        (throw d)
        v)
     (v (lambda (x ...) e) number #t #f - unspecified +
        call/cc)
     
     (P (store ((x v) ...) D d ...))
     (D (define x E) E)
     (E (v ... E e ...)
        (set! x E)
        (begin E e e ...)
        (if E e e)
        hole)
     
     (x (variable-except define beginD begin set! lambda call/cc + unspecified throw if))))
  
  (define lc-subst
    (subst
     [(? symbol?) (variable)]
     [(? number?) (constant)]
     [`(lambda (,@(xs ...)) ,b)
       (all-vars xs)
       (build (lambda (xs body) `(lambda ,xs ,body)))
       (subterm xs b)]
     [`(,f ,@(xs ...))
       (all-vars '())
       (build (lambda (vars f . xs) `(,f ,@xs)))
       (subterm '() f)
       (subterms '() xs)]))
  
  (define (subst-all vars args body)
    (let loop ([vars vars]
               [args args]
               [body body])
      (cond
        [(null? vars) body]
        [else (loop (cdr vars)
                    (cdr args)
                    (lc-subst (car vars) (car args) body))])))
  
  (define reductions
    (reduction-relation
     lang
     (--> (store ((x_s v_s) ...) (define x_1 v_1) d_1 ...)
          (store ((x_s v_s) ... (x_1 v_1)) unspecified d_1 ...)
          "TlDef"
          (side-condition (not (memq (term x_1) (term (x_s ...))))))
     
     (--> (store ((x_before v_before) ... (x_1 v_1) (x_after v_after) ...) (define x_1 v_2) d_1 ...)
          (store ((x_before v_before) ... (x_1 v_2) (x_after v_after) ...) unspecified d_1 ...)
          "TlReDef")
     
     (--> (store ((x_s v_s) ...) v_1 d_1 d_2 ...)
          (store ((x_s v_s) ...) d_1 d_2 ...)
          "TlToss")
     
     (--> (store ((x_s v_s) ...) (beginD d_1 ...) d_2 ...)
          (store ((x_s v_s) ...) d_1 ... d_2 ...)
          "TlBegin")
     
     (--> (store ((x_s v_s) ...) (in-hole D_1 (call/cc v_1)) d_1 ...)
          (store ((x_s v_s) ...) (in-hole D_1 (v_1 (lambda (x_2) (throw (in-hole D_1 x_2))))) d_1 ...)
          (fresh x_2)
          "TlCallcc")
     
     (--> (store ((x_s v_s) ...) ... (in-hole D_1 (throw d_1)) d_2 ...)
          (store ((x_s v_s) ...) ... d_1 d_2 ...)
          "TlThrow")
     
     (--> (store ((x_before v_before) ... (x_1 v_1) (x_after v_after) ...) (in-hole D_1 (set! x_1 v_2)) d_1 ...)
          (store ((x_before v_before) ... (x_1 v_2) (x_after v_after) ...) (in-hole D_1 unspecified) d_1 ...)
          "set!")
     (--> (side-condition (store ((x_s v_s) ...) (in-hole D_1 (set! x_1 v_1)) d_1 ...)
                          (not (memq (term x_1) (term (x_s ...)))))
          (store ((x_s v_s) ...) ,(format "attempt to set! free identifier: ~a" (term x_1)))
          
          "set! undefined")
     
     (--> (store ((x_before v_before) ... (x_1 v_1) (x_after v_after) ...) (in-hole D_1 x_1) d_1 ...)
          (store ((x_before v_before) ... (x_1 v_1) (x_after v_after) ...) (in-hole D_1 v_1) d_1 ...)
          "lookup")
     (--> (side-condition (store ((x_s v_s) ...) (in-hole D_1 x_1) d_1 ...)
                          (not (memq (term x_1) (term (x_s ...)))))
          (store ((x_s v_s) ...) ,(format "reference to free identifier: ~a" (term x_1)))
          "undefined")
     
     (--> (in-hole P_1 (begin v_1 e_1 e_2 ...)) 
          (in-hole P_1 (begin e_1 e_2 ...))
          "begin continue")
     (--> (in-hole P_1 (begin e_1))
          (in-hole P_1 e_1)
          "begin done")
     
     (--> (in-hole P_1 (if v_1 e_1 e_2)) 
          (in-hole P_1 e_1)
          (side-condition (not (equal? (term v_1) #f))))
     (--> (in-hole P_1 (if #f e_1 e_2)) 
          (in-hole P_1 e_2))
     
     (--> (in-hole P_1 ((lambda (x_1 ...) e_1) v_1 ...))
          (in-hole P_1 ,(subst-all (term (x_1 ...)) (term (v_1 ...)) (term e_1)))
          "beta-v")
     (--> (in-hole P_1 (+ number_1 ...))
          (in-hole P_1 ,(apply + (term (number_1 ...))))
          "+")))
  
  (provide reductions lang))
